import typer
from rich import print

pipeline_app = typer.Typer()

# Pipeline app
@pipeline_app.command('create')
def pipeline_create(
    name : str = typer.Option(..., help='pipeline name')
):
    '''
    Create a named pipeline

    Create a named pipeline. Pipeline is a list of scenarios
    to be run for specific cameras. The flow is as follows. Create
    a pipeline using:

    visionai pipeline create --name test_pipe

    visionai pipeline add-scenario --pipeline test_pipe  --name smoke-and-fire

    visionai pipeline add-scenario --pipeline test_pipe  --name ppe-detection

    visionai pipeline add-preprocess --pipeline test_pipe  --name face-blur

    visionai pipeline add-preprocess --pipeline test_pipe  --name text-blur

    visionai pipeline add-scenario --pipeline test_pipe  --name max-occupancy

    visionai pipeline show --pipeline test_pipe

    visionai pipeline add-camera --pipeline test_pipe  --name CAMERA-01

    visionai pipeline add-camera --pipeline test_pipe  --name CAMERA-02

    visionai pipeline show --pipeline test_pipe

    visionai pipeline run --pipeline test_pipe

    @arg pipeline - specify a named pipeline

    @return None
    '''
    print(f'Creating pipeline f{name}')


@pipeline_app.command('add-scenario')
def pipeline_add_scenario(
    pipeline : str = typer.Option(..., help='pipeline name'),
    scenario: str = typer.Option(..., help='scenario to add')
):
    '''
    Add a scenario to a pipeline

    The order of the scenarios does not matter. All added scenarios
    are run in different threads. All scenarios are run after
    pre-processing stage is done.

    $visionai pipeline --name test_pipe add-scenario --name smoke-and-fire

    $visionai pipeline --name test_pipe add-scenario --name ppe-detection

    $visionai pipeline --name test_pipe run

    @arg pipeline - specify a named pipeline
    @arg scenario - specify name of the scenario to run

    @return None
    '''
    print(f'Adding scenario {scenario} to pipeline f{pipeline}')


@pipeline_app.command('add-preprocess')
def pipeline_add_preprocess(
    pipeline : str = typer.Option(..., help='pipeline name'),
    preprocess: str = typer.Option(..., help='preprocess routine to add')
):
    '''
    Add a preprocess routine to a pipeline

    Preprocessing tasks are run prior to scenarios. The order in which
    multiple preprocess tasks are added does not matter. All added preprocess
    routines are executed in different threads.

    $ visionai pipeline --name test_pipe add-preprocess --name face-blur

    $ visionai pipeline --name test_pipe add-preprocess --name text-blur

    $ visionai pipeline --name test_pipe show

    $ visionai pipeline --name test_pipe run

    @arg pipeline - specify a named pipeline
    @arg preprocess - specify name of the preprocess task to run

    @return None
    '''
    print(f'Adding preprocess {preprocess} to pipeline f{pipeline}')


@pipeline_app.command('add-camera')
def pipeline_add_camera(
    pipeline : str = typer.Option(..., help='pipeline name'),
    camera: str = typer.Option(..., help='camera to add')
):
    '''
    Add a camera to a pipeline

    Each pipeline consists of a bunch of scenarios to run
    and which cameras they need to be run on. This method
    allows the user to add one or more named camera instance
    to a pipeline. Please note the camera instance has to be
    created prior to adding it here.

    # add a camera
    $ visionai camera add --name OFFICE-01 --uri https://youtube.com

    # add camera to pipeline
    $ visionai pipeline --name test_pipe add-camera --name OFFICE-01

    @arg pipeline - specify a named pipeline
    @arg camera - specify name of the camera to add

    @return None
    '''
    print(f'Adding camera {camera} to pipeline f{pipeline}')

@pipeline_app.command('remove-camera')
def pipeline_remove_camera(
    pipeline : str = typer.Option(..., help='pipeline name'),
    camera: str = typer.Option(..., help='camera to remove')
):
    '''
    Remove a camera from a pipeline

    This method can be used to remove a camera from a pipeline.

    $ visionai pipeline --name test_pipe remove-camera --name OFFICE-01

    @arg pipeline - specify a named pipeline
    @arg camera - specify name of the camera to remove

    @return None
    '''
    print(f'Removing camera {camera} to pipeline f{pipeline}')


@pipeline_app.command('reset')
def pipeline_reset(
    pipeline : str = typer.Option(..., help='pipeline name')
):
    '''
    Reset the pipeline to original state.

    Deletes all cameras, scenarios and scenario configuration
    from the pipeline. Its as if the pipeline has been deleted
    and created from scratch again.

    $ visionai pipeline --name test_pipe reset

    @arg pipeline - pipeline to reset

    @return None
    '''
    print(f'Pipeline {pipeline} reset')



@pipeline_app.command('show')
def pipeline_show(
    pipeline : str = typer.Option(..., help='pipeline name')
):
    '''
    Show details of a pipeline

    Show what is configured in the current pipeline.

    $ visionai pipeline --name test_pipe show

    @arg pipeline - specify a named pipeline

    @return None
    '''
    print(f'Show pipeline {pipeline} details')


@pipeline_app.command('run')
def run(
    pipeline: str = typer.Option(..., help='Pipeline to run')
):
    '''
    Run a pipeline of scenarios on given cameras

    Specify different scenarios to run on one or more
    cameras. This method can be directly used to specify scenarios
    and cameras directly. Else you can configure a named pipeline
    and then run it here.

    @arg pipeline - specify a named pipeline

    @return None
    '''
    print(f'Running pipeline f{pipeline}')

@pipeline_app.callback()
def callback():
    '''
    Manage pipelines

    Pipeline is a sequence of preprocess routines and scenarios
    to be run on a given set of cameras. Each pipeline can be
    configured to run specific scenarios - each scenario with their
    own customizations for event notifications. This module provides
    robust methods for managing pipelines, showing their details,
    adding/remove cameras from pipelines and running a pipeline.
   '''


if __name__ == '__main__':
    pipeline_app()



import os
import sys
from pathlib import Path
import unittest

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # visionai/visionai directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from util.general import WorkingDirectory, invoke_cmd

class TestDockerCli(unittest.TestCase):
    @WorkingDirectory(ROOT)
    def test_docker_cli(self):
        output = str(invoke_cmd(f'docker ps'))
        print(output)
        print('- - - - - - - - - - - - - - - - - - -')
        for idx, line in enumerate(output.split('\n')):
            print(f'{idx}: {line}')


if __name__ == '__main__':
    dockerObj = TestDockerCli()
    dockerObj.test_docker_cli()






import os
import sys
from pathlib import Path
import unittest

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # visionai/visionai directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from util.general import WorkingDirectory, invoke_cmd

class TestDeps(unittest.TestCase):
    @WorkingDirectory(ROOT)
    def test_import_torch(self):
        output = invoke_cmd(f'python -c "import torch; print(torch.__version__)"')
        assert output.startswith('1.1') # torch 1.12 above

    @WorkingDirectory(ROOT)
    def test_import_cv2(self):
        output = invoke_cmd(f'python -c "import cv2; print(cv2.__version__)"')
        assert output.startswith('4.') # cv2 4.6 or above

if __name__ == '__main__':
    unittest.main()




import os
import sys
import platform
import numpy as np
import cv2
import time
import torch
import torchvision
from PIL import Image

from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # edge-inference ROOT folder
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from util import TryExcept, emojis, WorkingDirectory
from util.general import git_describe, file_date, LOGGER, check_version
from util.metrics import box_iou


def smart_inference_mode(torch_1_9=check_version(torch.__version__, '1.9.0')):
    # Applies torch.inference_mode() decorator if torch>=1.9.0 else torch.no_grad() decorator
    def decorate(fn):
        return (torch.inference_mode if torch_1_9 else torch.no_grad)()(fn)

    return decorate

def select_device(device='', batch_size=0, newline=True):
    # device = None or 'cpu' or 0 or '0' or '0,1,2,3'
    s = f'YOLOv5 🚀 {git_describe() or file_date()} Python-{platform.python_version()} torch-{torch.__version__} '
    device = str(device).strip().lower().replace('cuda:', '').replace('none', '')  # to string, 'cuda:0' to '0'
    cpu = device == 'cpu'
    mps = device == 'mps'  # Apple Metal Performance Shaders (MPS)
    if cpu or mps:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # force torch.cuda.is_available() = False
    elif device:  # non-cpu device requested
        os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable - must be before assert is_available()
        assert torch.cuda.is_available() and torch.cuda.device_count() >= len(device.replace(',', '')), \
            f"Invalid CUDA '--device {device}' requested, use '--device cpu' or pass valid CUDA device(s)"

    if not cpu and not mps and torch.cuda.is_available():  # prefer GPU if available
        devices = device.split(',') if device else '0'  # range(torch.cuda.device_count())  # i.e. 0,1,6,7
        n = len(devices)  # device count
        if n > 1 and batch_size > 0:  # check batch_size is divisible by device_count
            assert batch_size % n == 0, f'batch-size {batch_size} not multiple of GPU count {n}'
        space = ' ' * (len(s) + 1)
        for i, d in enumerate(devices):
            p = torch.cuda.get_device_properties(i)
            s += f"{'' if i == 0 else space}CUDA:{d} ({p.name}, {p.total_memory / (1 << 20):.0f}MiB)\n"  # bytes to MB
        arg = 'cuda:0'
    elif mps and getattr(torch, 'has_mps', False) and torch.backends.mps.is_available():  # prefer MPS if available
        s += 'MPS\n'
        arg = 'mps'
    else:  # revert to CPU
        s += 'CPU\n'
        arg = 'cpu'

    if not newline:
        s = s.rstrip()
    LOGGER.info(s)
    return torch.device(arg)

def xyxy2xywh(x):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = (x[..., 0] + x[..., 2]) / 2  # x center
    y[..., 1] = (x[..., 1] + x[..., 3]) / 2  # y center
    y[..., 2] = x[..., 2] - x[..., 0]  # width
    y[..., 3] = x[..., 3] - x[..., 1]  # height
    return y


def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
    y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
    y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
    y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
    return y


def xywhn2xyxy(x, w=640, h=640, padw=0, padh=0):
    # Convert nx4 boxes from [x, y, w, h] normalized to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = w * (x[..., 0] - x[..., 2] / 2) + padw  # top left x
    y[..., 1] = h * (x[..., 1] - x[..., 3] / 2) + padh  # top left y
    y[..., 2] = w * (x[..., 0] + x[..., 2] / 2) + padw  # bottom right x
    y[..., 3] = h * (x[..., 1] + x[..., 3] / 2) + padh  # bottom right y
    return y


def xyxy2xywhn(x, w=640, h=640, clip=False, eps=0.0):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] normalized where xy1=top-left, xy2=bottom-right
    if clip:
        clip_boxes(x, (h - eps, w - eps))  # warning: inplace clip
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = ((x[..., 0] + x[..., 2]) / 2) / w  # x center
    y[..., 1] = ((x[..., 1] + x[..., 3]) / 2) / h  # y center
    y[..., 2] = (x[..., 2] - x[..., 0]) / w  # width
    y[..., 3] = (x[..., 3] - x[..., 1]) / h  # height
    return y


def xyn2xy(x, w=640, h=640, padw=0, padh=0):
    # Convert normalized segments into pixel segments, shape (n,2)
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = w * x[..., 0] + padw  # top left x
    y[..., 1] = h * x[..., 1] + padh  # top left y
    return y



def letterbox(im, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, stride=32):

    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, ratio, (dw, dh)

def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
    y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
    y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
    y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
    return y

def non_max_suppression(
        prediction,
        conf_thres=0.25,
        iou_thres=0.45,
        classes=None,
        agnostic=False,
        multi_label=False,
        labels=(),
        max_det=300,
        nm=0,  # number of masks
):
    """Non-Maximum Suppression (NMS) on inference results to reject overlapping detections
    Returns:
         list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """

    # Checks
    assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
    assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'
    if isinstance(prediction, (list, tuple)):  # YOLOv5 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output

    device = prediction.device
    mps = 'mps' in device.type  # Apple MPS
    if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
        prediction = prediction.cpu()
    bs = prediction.shape[0]  # batch size
    nc = prediction.shape[2] - nm - 5  # number of classes
    xc = prediction[..., 4] > conf_thres  # candidates

    # Settings
    # min_wh = 2  # (pixels) minimum box width and height
    max_wh = 7680  # (pixels) maximum box width and height
    max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
    time_limit = 0.5 + 0.05 * bs  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    mi = 5 + nc  # mask start index
    output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            lb = labels[xi]
            v = torch.zeros((len(lb), nc + nm + 5), device=x.device)
            v[:, :4] = lb[:, 1:5]  # box
            v[:, 4] = 1.0  # conf
            v[range(len(lb)), lb[:, 0].long() + 5] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

        # Box/Mask
        box = xywh2xyxy(x[:, :4])  # center_x, center_y, width, height) to (x1, y1, x2, y2)
        mask = x[:, mi:]  # zero columns if no masks

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:mi] > conf_thres).nonzero(as_tuple=False).T
            x = torch.cat((box[i], x[i, 5 + j, None], j[:, None].float(), mask[i]), 1)
        else:  # best class only
            conf, j = x[:, 5:mi].max(1, keepdim=True)
            x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence and remove excess boxes

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
        i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        i = i[:max_det]  # limit detections
        if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
            # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
            iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
            weights = iou * scores[None]  # box weights
            x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
            if redundant:
                i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if mps:
            output[xi] = output[xi].to(device)
        if (time.time() - t) > time_limit:
            LOGGER.warning(f'WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded')
            break  # time limit exceeded

    return output


def exif_transpose(image):
    """
    Transpose a PIL image accordingly if it has an EXIF Orientation tag.
    Inplace version of https://github.com/python-pillow/Pillow/blob/master/src/PIL/ImageOps.py exif_transpose()

    :param image: The image to transpose.
    :return: An image.
    """
    exif = image.getexif()
    orientation = exif.get(0x0112, 1)  # default 1
    if orientation > 1:
        method = {
            2: Image.FLIP_LEFT_RIGHT,
            3: Image.ROTATE_180,
            4: Image.FLIP_TOP_BOTTOM,
            5: Image.TRANSPOSE,
            6: Image.ROTATE_270,
            7: Image.TRANSVERSE,
            8: Image.ROTATE_90}.get(orientation)
        if method is not None:
            image = image.transpose(method)
            del exif[0x0112]
            image.info["exif"] = exif.tobytes()
    return image


def clip_boxes(boxes, shape):
    # Clip boxes (xyxy) to image shape (height, width)
    if isinstance(boxes, torch.Tensor):  # faster individually
        boxes[..., 0].clamp_(0, shape[1])  # x1
        boxes[..., 1].clamp_(0, shape[0])  # y1
        boxes[..., 2].clamp_(0, shape[1])  # x2
        boxes[..., 3].clamp_(0, shape[0])  # y2
    else:  # np.array (faster grouped)
        boxes[..., [0, 2]] = boxes[..., [0, 2]].clip(0, shape[1])  # x1, x2
        boxes[..., [1, 3]] = boxes[..., [1, 3]].clip(0, shape[0])  # y1, y2


def clip_segments(segments, shape):
    # Clip segments (xy1,xy2,...) to image shape (height, width)
    if isinstance(segments, torch.Tensor):  # faster individually
        segments[:, 0].clamp_(0, shape[1])  # x
        segments[:, 1].clamp_(0, shape[0])  # y
    else:  # np.array (faster grouped)
        segments[:, 0] = segments[:, 0].clip(0, shape[1])  # x
        segments[:, 1] = segments[:, 1].clip(0, shape[0])  # y


def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None):
    # Rescale boxes (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    boxes[..., [0, 2]] -= pad[0]  # x padding
    boxes[..., [1, 3]] -= pad[1]  # y padding
    boxes[..., :4] /= gain
    clip_boxes(boxes, img0_shape)
    return boxes




import requests
import json
from pathlib import Path

FILE = Path(__file__).resolve()
CONFIGDIR = FILE.parents[0]  # config directory

from config import SCENARIOS_URL, SCENARIOS_OVERRIDE

def main():
    res = requests.get(SCENARIOS_URL)
    scenarios = res.json()
    with open(SCENARIOS_OVERRIDE, 'w') as f:
        json.dump(scenarios, f, indent=4)

    print(f'Scenarios schema {SCENARIOS_URL} downloaded to {SCENARIOS_OVERRIDE}')

if __name__ == '__main__':
    main()



import os
import sys
from pathlib import Path

FILE = Path(__file__).resolve()
SCENARIOS = FILE.parents[0]
ROOT = FILE.parents[1]  # Root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
ROOT = Path(os.path.relpath(ROOT, Path.cwd()))  # relative

def yolov5_triton(url, model_name, labels=None):

    from models.common import AutoShape, Yolov5Triton
    from util.general import LOGGER, check_requirements
    from util.image_utils import select_device

    device = select_device()
    LOGGER.info(f'Using {url} as Triton inference server. Model name {model_name}')
    check_requirements('tritonclient[all]', install=True)

    model = Yolov5Triton(url=url, model_name=model_name, labels=labels)
    model = AutoShape(model)    # Add nms, auto-scaling results
    return model.to(device)

if __name__ == '__main__':

    import cv2
    import numpy as np
    from PIL import Image
    from util.general import WorkingDirectory

    # Model
    model_smoke = yolov5_triton(url='http://localhost:8000', model_name='smoke-and-fire-detection')
    # model_ppe = yolov5_triton(url='http://localhost:8000', model_name='ppe-detection')
    # model_yolov5s = yolov5_triton(url='http://localhost:8000', model_name='yolov5s')

    # or use grpc
    # model = yolov5_triton(url='grpc://0.0.0.0:8001', model='ppe-detection')

    with WorkingDirectory(SCENARIOS):
        # Images
        imgs = [
            'data/images/zidane.jpg',  # filename
            Path('data/images/zidane.jpg'),  # Path
            'https://ultralytics.com/images/zidane.jpg',  # URI
            # cv2.imread('data/images/bus.jpg')[:, :, ::-1],  # OpenCV
            # Image.open('data/images/person-helmet.jpg'),  # PIL
            np.zeros((320, 640, 3))]  # numpy

        # Inference
        for idx, img in enumerate(imgs):
            results = model_smoke(img, size=640)  # batched inference
            results.print()
            results.show()
            # results.save()

            # results = model_ppe(img, size=640)  # batched inference
            # results.print()
            # results.save()

            # results = model_yolov5s(img, size=640)  # batched inference
            # results.print()
            # results.save()



import os
import sys
import re
import json
import argparse
from urllib.parse import urlparse
import shutil

from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # ROOT folder
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))
ROOT = Path(os.path.relpath(ROOT, Path.cwd()))

from util.general import LOGGER, print_args, check_requirements, file_md5, colorstr, invoke_cmd, WorkingDirectory

def safe_download_to_file(file, url, url2=None, min_bytes=1E0, error_msg=''):
    # Attempts to download file from url or url2, checks and removes incomplete downloads < min_bytes
    from util.general import LOGGER
    import torch

    file = Path(file)
    assert_msg = f"Downloaded file '{file}' does not exist or size is < min_bytes={min_bytes}"
    try:  # url1
        torch.hub.download_url_to_file(url, str(file), progress=True)
        assert file.exists() and file.stat().st_size > min_bytes, assert_msg  # check
    except Exception as e:  # url2
        if file.exists():
            file.unlink()  # remove partial downloads
        LOGGER.info(f'ERROR: {e}\nRe-attempting {url2 or url} to {file}...')
        os.system(f"curl -# -L '{url2 or url}' -o '{file}' --retry 3 -C -")  # curl download, retry and resume on fail
    finally:
        if not file.exists() or file.stat().st_size < min_bytes:  # check
            if file.exists():
                file.unlink()  # remove partial downloads
            LOGGER.info(f"ERROR: {assert_msg}\n{error_msg}")
        LOGGER.info('')

def safe_download_to_folder(
    url=None,
    dirname='.',
    overwrite=False
):
    # url & path check
    if url is None or os.path.isdir(dirname) is False:
        LOGGER.error('URL ({}) or dirname ({}) not valid'.format(url, dirname))
        return False

    # Find the path to save the model file.
    file_name = os.path.basename(urlparse(url).path)
    file_path = Path(dirname) / file_name

    # If the file exists, then skip download
    if overwrite is False and os.path.exists(file_path):
        LOGGER.info(f'{file_path} already exists, skip download.')
        return False

    # Download the URL to path
    safe_download_to_file(file_path, url)

    # If its a zip file, unzip it as well.
    if file_path.suffix == '.zip':
        shutil.unpack_archive(file_path, dirname) # unzip
        # os.remove(file_path)   # remove zip file

    return True

def download_models(
    scenarios=[],
    check_md5=False
):
    if scenarios is None or len(scenarios) == 0:
        LOGGER.error('Scenarios not specified')
        return False

    for scen in scenarios:
        model_url = scen['model_url']

        # Nothing to download
        if model_url is None:
            continue

        scen_name = scen['name']
        model_name = os.path.basename(urlparse(model_url).path)
        model_hash = scen['model_hash']
        model_folder = ROOT / 'run'
        model_file_path = Path(model_folder) / model_name

        # Check if model file exists before downloading.
        redownload = False
        if check_md5 and os.path.exists(model_file_path) and model_hash != file_md5(model_file_path):
            redownload = True
            LOGGER.warning(f'⛔ {colorstr("red", "bold", scen_name)}: Model MD5 has changed')

        # If not present or redownload, download file.
        if not os.path.exists(model_file_path) or redownload:
            LOGGER.info(f'{colorstr(model_name)}: Downloading {model_url}')
            safe_download_to_folder(url=model_url, dirname=model_folder)
            if check_md5 and model_hash != file_md5(model_file_path):
                LOGGER.error(f'⛔ {colorstr("red", "bold", model_name)}: Downloaded model file, but MD5 doesn\'t match')

def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument('--scenarios', type=str, default='scenarios.json', help='Scenarios file.json')
    parser.add_argument('--url', type=str, default=None, help='Model url (ex: https://workplaceos.blob.core.windows.net/models/ppe-detection/ppe-detection-0.0.1.pt)')
    parser.add_argument('--dirname', type=str, default='.', help='Results folder')
    parser.add_argument('--verify', action='store_true', help='Validate MD5 for downloaded files')
    opt = parser.parse_args()
    print_args(vars(opt))
    return opt

def main(opt):
    check_requirements()
    if opt.scenarios is not None:
        with open(opt.scenarios, 'r') as f:
            data = json.load(f)
        scenarios = data['scenarios']
        download_models(scenarios, opt.verify)
    else:
        safe_download_to_folder(opt.url, opt.dirname)

# Main method.
if __name__ == '__main__':
    opt = parse_opt()
    main(opt)



import typer
from rich import print
import docker
from rich.console import Console
from rich import print

import sys
from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # visionai/visionai directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from config import *
from util.docker_utils import *

err_console = Console(stderr=True)
web_app = typer.Typer()

@web_app.command('start')
def web_start():
    '''
    Start web server

    Use this function to start the web-service. Web service
    can be used for more intuitive configuration for the
    cameras and scenarios. Web-app is also the place to view
    event details, camera live-stream etc.
    '''
    try:
        print('- - - - - - - - - - - - - - - - - - - - - - - - - - -')

        # Check if already running
        web_app_running = False
        web_api_running = False
        redis_running = False
        grafana_running = False

        client = docker.from_env()
        containers = client.containers.list()
        for container in containers:
            if container.name == WEB_APP_CONTAINER_NAME:
                print(f'Web app available at: http://localhost:{WEB_APP_PORT}')
                web_app_running = True
            if container.name == WEB_API_CONTAINER_NAME:
                print(f'Web service API available at: http://localhost:{WEB_API_PORT}')
                web_api_running = True
            if container.name == REDIS_CONTAINER_NAME:
                print(f'Redis server is at: redis://localhost:{REDIS_SERVER_PORT}')
                redis_running = True
            if container.name == GRAFANA_CONTAINER_NAME:
                print(f'Grafana server is at: http://localhost:{GRAFANA_SERVER_PORT}')
                grafana_running = True

        # If all services are running, return
        if web_app_running and web_api_running and redis_running and grafana_running:
            return

        if web_app_running is False:
            print(f'Starting web app at port {WEB_APP_PORT}')
            docker_container_start(
                container_name=WEB_APP_CONTAINER_NAME,
                image=WEB_APP_DOCKER_IMAGE,
                portmap={80:WEB_APP_PORT},
                network_name=DOCKER_NETWORK
            )
            print(f'Web app available at: http://localhost:{WEB_APP_PORT}')

        if web_api_running is False:
            print(f'Starting web service API at port {WEB_API_PORT}')

            if sys.platform == 'win32':
                DOCKER_SOCK = '//var/run/docker.sock'
            else:
                DOCKER_SOCK = '/var/run/docker.sock'

            docker_container_start(
                container_name=WEB_API_CONTAINER_NAME,
                image=WEB_API_DOCKER_IMAGE,
                portmap={3002:WEB_API_PORT},
                network_name=DOCKER_NETWORK,
                volmap=[
                    f'{WEB_API_MODELS_REPO}:/models',
                    f'{WEB_API_CONFIG_FOLDER}:/config',
                    f'{DOCKER_SOCK}:/var/run/docker.sock'
                ],
                command='python server.py --models-repo /models --config /config'
            )
            print(f'Web service API available at: http://localhost:{WEB_API_PORT}')

        if redis_running is False:
            print(f'Starting redis server at port {REDIS_SERVER_PORT}')
            redis_container_start()
            redis_python_install()
            print(f'Redis server is at: redis://localhost:{REDIS_SERVER_PORT}')

        if grafana_running is False:
            print(f'Starting grafana server at port {GRAFANA_SERVER_PORT}')
            grafana_container_start()
            print(f'Grafana server is at: http://localhost:{GRAFANA_SERVER_PORT}')

    except Exception as e:
        err_console.print_exception()
        print(f'Error: {e}')

@web_app.command('stop')
def web_stop(web: str=None):
    '''
    Stop web server

    Use this function to stop already running web-service. There
    can be a single instance of the web-service supported currently.
    So there is no need for any arguments for this function.
    '''
    try:
        print('- - - - - - - - - - - - - - - - - - - - - - - - - - -')
        print(f'Stop web-app....')
        client = docker.from_env()
        web_service_container = client.containers.get(WEB_APP_CONTAINER_NAME)
        web_service_container.stop()
    except docker.errors.NotFound :
        message = typer.style(f"Web-app not running", fg=typer.colors.WHITE, bg=typer.colors.RED)
        typer.echo(message)

    try:
        print(f'Stop API service....')
        web_api_container = client.containers.get(WEB_API_CONTAINER_NAME)
        web_api_container.stop()
    except docker.errors.NotFound :
        message = typer.style(f"Web-API not running", fg=typer.colors.WHITE, bg=typer.colors.RED)
        typer.echo(message)

    try:
        print(f'Stop redis server....')
        redis_container = client.containers.get(REDIS_CONTAINER_NAME)
        redis_container.stop()
    except docker.errors.NotFound :
        message = typer.style(f"Redis not running", fg=typer.colors.WHITE, bg=typer.colors.RED)
        typer.echo(message)

    try:
        print(f'Stop grafana server....')
        grafana_container = client.containers.get(GRAFANA_CONTAINER_NAME)
        grafana_container.stop()
    except docker.errors.NotFound :
        message = typer.style(f"Grafana not running", fg=typer.colors.WHITE, bg=typer.colors.RED)
        typer.echo(message)

        print(f'Done.')

def print_container_status(container_name, tail):
    try:
        print('- - - - - - - - - - - - - - - - - - - - - - - - - - -')
        print(f'{container_name} status....')
        client = docker.from_env()
        ctainer = client.containers.get(container_name)
        ctainer_status = typer.style(ctainer.status, fg=typer.colors.WHITE, bg=typer.colors.GREEN)
        typer.echo(ctainer_status)
        logs = ctainer.logs(tail=tail)
        log_message= logs.decode("utf-8")
        print(log_message)
        web_service_port_message = typer.style(ctainer.ports, fg=typer.colors.WHITE, bg=typer.colors.GREEN)
        typer.echo(web_service_port_message)
        print('- - - - - - - - - - - - - - - - - - - - - - - - - - -')

    except docker.errors.NotFound:
        message = typer.style(f"{container_name} not running", fg=typer.colors.WHITE, bg=typer.colors.RED)
        typer.echo(message)

@web_app.command('status')
def web_status(
    tail: int = typer.Option(20, help='tail number of lines')
    ):
    '''
    Web service status

    Use this function to get the status of the web-service. (if
    its running or not. This function also prints diagnostic
    information like last few log messages etc.)
    '''
    print_container_status(WEB_APP_CONTAINER_NAME, tail)
    print_container_status(WEB_API_CONTAINER_NAME, tail)
    print_container_status(REDIS_CONTAINER_NAME, tail)
    print_container_status(GRAFANA_CONTAINER_NAME, tail)


@web_app.callback()
def callback():
    '''
    Start/stop web-app

    Start and stop the VisionAI web-app which can be a more
    intuitive way of managing cameras, pipelines and scenarios.
    Web-app also provides a live-stream view of the cameras.
    '''
    pass

if __name__ == '__main__':
    # web_app()

    web_stop()


from rich import print
import time

import sys
import cv2
import time

from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # visionai/visionai directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from util.general import LOGGER
from scenarios import Scenario
from config import TRITON_HTTP_URL
from util.track import *

from events.events_engine import EventsEngine
from enum import Enum

class Event(str, Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'

class OccupancyMonitoring(Scenario):
    alert_time = 10
    def __init__(self, scenario_name, camera_name=0, events=None, triton_url=TRITON_HTTP_URL):

        from models.triton_client_yolov5 import yolov5_triton
        self.model = yolov5_triton(triton_url, scenario_name)
        self.f_event =  EventsEngine(use_redis=False)
        self.line_thickness=3
        super().__init__(scenario_name, camera_name, events, triton_url)


    def draw_boxes(self, img, bbox, count, total_time, offset=(0, 0)):
        for i, box in enumerate(bbox):
            x1, y1, x2, y2 = [int(i) for i in box]
            x1 += offset[0]
            x2 += offset[0]
            y1 += offset[1]
            y2 += offset[1]
            data = (int((box[0]+box[2])/2),(int((box[1]+box[3])/2)))
            label = f"ID: {str(count)} - Duration: {str(int(total_time))}"

            (w, h), _ = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.6, 1)
            cv2.rectangle(img, (x1, y1), (x2, y2),(255,191,0), 2)
            cv2.rectangle(img, (x1, y1 - 20), (x1 + w, y1), (255,191,0), -1)
            cv2.putText(img, label, (x1, y1 - 5),cv2.FONT_HERSHEY_SIMPLEX, 0.6, 
            [255, 255, 255], 1)
            cv2.circle(img, data, 3, (255,191,0),-1)
        return img


    def start(self, camera_name=0):
        '''
        Stream processing

        When running a scenario - the caller can specify any specific camera.
        '''

        stream = camera_name

        print(f'Opening capture for {stream}')
        video = cv2.VideoCapture(stream)
        info_dic = {}
        sort_max_age = 5 
        sort_min_hits = 2
        sort_iou_thresh = 0.2
        sort_tracker = Track(max_age=sort_max_age,
                        min_hits=sort_min_hits,
                        iou_threshold=sort_iou_thresh) 

        prev_frame_time = time.time()
        new_frame_time = 0
        while True:
            # Do processing
            ret, frame = video.read()
            # fps = video.get(cv2.CAP_PROP_FPS)
            if ret is False:
                LOGGER.error('ERROR: reading from video frame')
                time.sleep(1)
                continue

            new_frame_time = time.time()
            fps = 1/(new_frame_time-prev_frame_time)
            # total_time += (new_frame_time-prev_frame_time)

            prev_frame_time = new_frame_time

            # Detect smoke & fire
            results = self.model(frame, size=640)  # batched inference

            det = results.xyxy[0]
            print("-----------------------------------")
            if len(det):
                dets_to_sort = np.empty((0,6))
                for x1,y1,x2,y2,conf,detclass in det.cpu().detach().numpy():
                    if int(detclass) == 0:
                        dets_to_sort = np.vstack((dets_to_sort, 
                                                np.array([x1, y1, x2, y2, 
                                                            conf, detclass])))

                tracked_dets = sort_tracker.update(dets_to_sort)

                if len(tracked_dets)>0:
                    bbox_xyxy = tracked_dets[:,:4]
                    identities = tracked_dets[:, 8]

                    for i in range(len(identities)):
                        id = identities[i]
                        bbox = bbox_xyxy[i]
                        if id in info_dic:
                            info_dic[id] = {
                                "id": int(id),
                                "count": info_dic[id]["count"] + 1,
                                "frames": info_dic[id]["frames"] + 1, 
                                "duration": round((info_dic[id]["duration"] + 1/fps), 2),
                                "bbox": list(bbox)
                            } 
                            self.draw_boxes(frame, [bbox], info_dic[id]['id'], info_dic[id]['duration'])  
                            if info_dic[id]["duration"] > self.alert_time:
                                self.f_event.fire_event(Event.CRITICAL, 'WEBCAM', "occupancy-monitoring", 'TIME-EXCEEDED', info_dic[id])         

                        else:
                            info_dic[id] = {"id": int(id), "count": 1, "frames": 1, "duration": 0, "bbox": list(bbox)}

                            self.draw_boxes(frame, [bbox], info_dic[id]['id'], info_dic[id]['duration'])                 
                                
                cv2.imshow('Output', frame)
                key = cv2.waitKey(5)
                if key == 27:
                    import sys
                    print('Exiting.')
                    sys.exit(0)
                        
            results.print()
            if self.stop_evt.is_set():
                break

def camera_stream():
    snf = OccupancyMonitoring(scenario_name = 'occupancy-monitoring')
    snf.start(camera_name=0)

if __name__ == '__main__':
    camera_stream()


from __future__ import print_function

import numpy as np

from filterpy.kalman import KalmanFilter

np.random.seed(0)

def linear_assignment(cost_matrix):
    try:
        import lap #linear assignment problem solver
        _, x, y = lap.lapjv(cost_matrix, extend_cost = True)
        return np.array([[y[i],i] for i in x if i>=0])
    except ImportError:
        from scipy.optimize import linear_sum_assignment
        x,y = linear_sum_assignment(cost_matrix)
        return np.array(list(zip(x,y)))


"""From SORT: Computes IOU between two boxes in the form [x1,y1,x2,y2]"""
def iou_batch(bb_test, bb_gt):
    
    bb_gt = np.expand_dims(bb_gt, 0)
    bb_test = np.expand_dims(bb_test, 1)
    
    xx1 = np.maximum(bb_test[...,0], bb_gt[..., 0])
    yy1 = np.maximum(bb_test[..., 1], bb_gt[..., 1])
    xx2 = np.minimum(bb_test[..., 2], bb_gt[..., 2])
    yy2 = np.minimum(bb_test[..., 3], bb_gt[..., 3])
    w = np.maximum(0., xx2 - xx1)
    h = np.maximum(0., yy2 - yy1)
    wh = w * h
    o = wh / ((bb_test[..., 2] - bb_test[..., 0]) * (bb_test[..., 3] - bb_test[..., 1])                                      
    + (bb_gt[..., 2] - bb_gt[..., 0]) * (bb_gt[..., 3] - bb_gt[..., 1]) - wh)
    return(o)


"""Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form [x,y,s,r] where x,y is the center of the box and s is the scale/area and r is the aspect ratio"""
def convert_bbox_to_z(bbox):
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0] + w/2.
    y = bbox[1] + h/2.
    s = w * h    
    #scale is just area
    r = w / float(h)
    return np.array([x, y, s, r]).reshape((4, 1))


"""Takes a bounding box in the centre form [x,y,s,r] and returns it in the form
    [x1,y1,x2,y2] where x1,y1 is the top left and x2,y2 is the bottom right"""
def convert_x_to_bbox(x, score=None):
    w = np.sqrt(x[2] * x[3])
    h = x[2] / w
    if(score==None):
        return np.array([x[0]-w/2.,x[1]-h/2.,x[0]+w/2.,x[1]+h/2.]).reshape((1,4))
    else:
        return np.array([x[0]-w/2.,x[1]-h/2.,x[0]+w/2.,x[1]+h/2.,score]).reshape((1,5))

"""This class represents the internal state of individual tracked objects observed as bbox."""
class KalmanBoxTracker(object):
    
    count = 0
    def __init__(self, bbox):
        """
        Initialize a tracker using initial bounding box
        
        Parameter 'bbox' must have 'detected class' int number at the -1 position.
        """
        self.kf = KalmanFilter(dim_x=7, dim_z=4)
        self.kf.F = np.array([[1,0,0,0,1,0,0],[0,1,0,0,0,1,0],[0,0,1,0,0,0,1],[0,0,0,1,0,0,0],[0,0,0,0,1,0,0],[0,0,0,0,0,1,0],[0,0,0,0,0,0,1]])
        self.kf.H = np.array([[1,0,0,0,0,0,0],[0,1,0,0,0,0,0],[0,0,1,0,0,0,0],[0,0,0,1,0,0,0]])

        self.kf.R[2:,2:] *= 10. # R: Covariance matrix of measurement noise (set to high for noisy inputs -> more 'inertia' of boxes')
        self.kf.P[4:,4:] *= 1000. #give high uncertainty to the unobservable initial velocities
        self.kf.P *= 10.
        self.kf.Q[-1,-1] *= 0.5 # Q: Covariance matrix of process noise (set to high for erratically moving things)
        self.kf.Q[4:,4:] *= 0.5

        self.kf.x[:4] = convert_bbox_to_z(bbox) # STATE VECTOR
        self.time_since_update = 0
        self.id = KalmanBoxTracker.count
        KalmanBoxTracker.count += 1
        self.history = []
        self.hits = 0
        self.hit_streak = 0
        self.age = 0
        self.centroidarr = []
        CX = (bbox[0]+bbox[2])//2
        CY = (bbox[1]+bbox[3])//2
        self.centroidarr.append((CX,CY))
        
        
        #keep yolov5 detected class information
        self.detclass = bbox[5]
        
    def update(self, bbox):
        """
        Updates the state vector with observed bbox
        """
        self.time_since_update = 0
        self.history = []
        self.hits += 1
        self.hit_streak += 1
        self.kf.update(convert_bbox_to_z(bbox))
        self.detclass = bbox[5]
        CX = (bbox[0]+bbox[2])//2
        CY = (bbox[1]+bbox[3])//2
        self.centroidarr.append((CX,CY))
        
    def predict(self):
        """
        Advances the state vector and returns the predicted bounding box estimate
        """
        if((self.kf.x[6]+self.kf.x[2])<=0):
            self.kf.x[6] *= 0.0
        self.kf.predict()
        self.age += 1
        if(self.time_since_update>0):
            self.hit_streak = 0
        self.time_since_update += 1
        self.history.append(convert_x_to_bbox(self.kf.x))
        # bbox=self.history[-1]
        # CX = (bbox[0]+bbox[2])/2
        # CY = (bbox[1]+bbox[3])/2
        # self.centroidarr.append((CX,CY))
        
        return self.history[-1]
    
    
    def get_state(self):
        """
        Returns the current bounding box estimate
        # test
        arr1 = np.array([[1,2,3,4]])
        arr2 = np.array([0])
        arr3 = np.expand_dims(arr2, 0)
        np.concatenate((arr1,arr3), axis=1)
        """
        arr_detclass = np.expand_dims(np.array([self.detclass]), 0)
        
        arr_u_dot = np.expand_dims(self.kf.x[4],0)
        arr_v_dot = np.expand_dims(self.kf.x[5],0)
        arr_s_dot = np.expand_dims(self.kf.x[6],0)
        
        return np.concatenate((convert_x_to_bbox(self.kf.x), arr_detclass, arr_u_dot, arr_v_dot, arr_s_dot), axis=1)
    
def associate_detections_to_trackers(detections, trackers, iou_threshold = 0.3):
    """
    Assigns detections to tracked object (both represented as bounding boxes)
    Returns 3 lists of 
    1. matches,
    2. unmatched_detections
    3. unmatched_trackers
    """
    if(len(trackers)==0):
        return np.empty((0,2),dtype=int), np.arange(len(detections)), np.empty((0,5),dtype=int)
    
    iou_matrix = iou_batch(detections, trackers)
    
    if min(iou_matrix.shape) > 0:
        a = (iou_matrix > iou_threshold).astype(np.int32)
        if a.sum(1).max() == 1 and a.sum(0).max() ==1:
            matched_indices = np.stack(np.where(a), axis=1)
        else:
            matched_indices = linear_assignment(-iou_matrix)
    else:
        matched_indices = np.empty(shape=(0,2))
    
    unmatched_detections = []
    for d, det in enumerate(detections):
        if(d not in matched_indices[:,0]):
            unmatched_detections.append(d)
    
    unmatched_trackers = []
    for t, trk in enumerate(trackers):
        if(t not in matched_indices[:,1]):
            unmatched_trackers.append(t)
    
    #filter out matched with low IOU
    matches = []
    for m in matched_indices:
        if(iou_matrix[m[0], m[1]]<iou_threshold):
            unmatched_detections.append(m[0])
            unmatched_trackers.append(m[1])
        else:
            matches.append(m.reshape(1,2))
    
    if(len(matches)==0):
        matches = np.empty((0,2), dtype=int)
    else:
        matches = np.concatenate(matches, axis=0)
        
    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)
    

class Track(object):
    def __init__(self, max_age=1, min_hits=3, iou_threshold=0.3):
        """
        Parameters for SORT
        """
        self.max_age = max_age
        self.min_hits = min_hits
        self.iou_threshold = iou_threshold
        self.trackers = []
        self.frame_count = 0
    def getTrackers(self,):
        return self.trackers
        
    def update(self, dets= np.empty((0,6))):
        """
        Parameters:
        'dets' - a numpy array of detection in the format [[x1, y1, x2, y2, score], [x1,y1,x2,y2,score],...]
        
        Ensure to call this method even frame has no detections. (pass np.empty((0,5)))
        
        Returns a similar array, where the last column is object ID (replacing confidence score)
        
        NOTE: The number of objects returned may differ from the number of objects provided.
        """
        self.frame_count += 1
        
        # Get predicted locations from existing trackers
        trks = np.zeros((len(self.trackers), 6))
        to_del = []
        ret = []
        for t, trk in enumerate(trks):
            pos = self.trackers[t].predict()[0]
            trk[:] = [pos[0], pos[1], pos[2], pos[3], 0, 0]
            if np.any(np.isnan(pos)):
                to_del.append(t)
        trks = np.ma.compress_rows(np.ma.masked_invalid(trks))
        for t in reversed(to_del):
            self.trackers.pop(t)
        matched, unmatched_dets, unmatched_trks = associate_detections_to_trackers(dets, trks, self.iou_threshold)
        
        # Update matched trackers with assigned detections
        for m in matched:
            self.trackers[m[1]].update(dets[m[0], :])
            
        # Create and initialize new trackers for unmatched detections
        for i in unmatched_dets:
            trk = KalmanBoxTracker(np.hstack((dets[i,:], np.array([0]))))
            #trk = KalmanBoxTracker(np.hstack(dets[i,:])
            self.trackers.append(trk)
        
        i = len(self.trackers)
        for trk in reversed(self.trackers):
            d = trk.get_state()[0]
            if (trk.time_since_update < 1) and (trk.hit_streak >= self.min_hits or self.frame_count <= self.min_hits):
                ret.append(np.concatenate((d, [trk.id+1])).reshape(1,-1)) #+1'd because MOT benchmark requires positive value
            i -= 1
            #remove dead tracklet
            if(trk.time_since_update >self.max_age):
                self.trackers.pop(i)
        if(len(ret) > 0):
            return np.concatenate(ret)
        return np.empty((0,6))




